#! /usr/bin/env python3
#
#  Copyright 2019 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import scipy
from spectral.io import envi
from skimage.segmentation import slic
import numpy as np
import ray
import ray.services
import atexit
import logging

@ray.remote
def segment_chunk(lstart, lend, in_file, nodata_value, npca, segsize, logfile=None, loglevel='INFO'):
    """
    Segment a small chunk of the image

    Args:
        lstart: starting position in image file
        lend:  stopping position in image file
        in_file: file path to segment
        nodata_value: value to ignore
        npca:  number of pca components to use
        segsize: mean segmentation size
        logfile: logging file name
        loglevel: logging level

    Returns:
        lstart: starting position in image file
        lend: stopping position in image file
        labels: labeled image chunk

    """
    logging.basicConfig(format='%(levelname)s:%(message)s', level=loglevel, filename=logfile)

    logging.info(f'{lstart}: starting')

    in_img = envi.open(in_file + '.hdr', in_file)
    meta = in_img.metadata
    nl, nb, ns = [int(meta[n]) for n in ('lines', 'bands', 'samples')]
    img_mm = in_img.open_memmap(interleave='bip', writable=False)

    # Do quick single-band screen before reading all bands
    use = np.logical_not(np.isclose(np.array(img_mm[lstart:lend, :, 0]), nodata_value))
    if np.sum(use) == 0:
        logging.info(f'{lstart}: no non null data present, returning early')
        return lstart, lend, np.zeros((use.shape[0],ns))


    x = np.array(img_mm[lstart:lend, :, :]).astype(np.float32)
    nc = x.shape[0]
    x = x.reshape((nc * ns, nb))
    logging.debug(f'{lstart}: read and reshaped data')

    # Excluding bad locations, calculate top PCA coefficients
    use = np.all(abs(x - nodata_value) > 1e-6, axis=1)

    # If this chunk is empty, return immediately
    if np.sum(use) == 0:
        logging.info(f'{lstart}: no non null data present, returning early')
        return lstart, lend, np.zeros((nc,ns))

    mu = x[use, :].mean(axis=0)
    C = np.cov(x[use, :], rowvar=False)
    [v, d] = scipy.linalg.eigh(C)

    # Determine segmentation compactness scaling based on eigenvalues
    # Override with a floor value to prevent zeros
    cmpct = scipy.linalg.norm(np.sqrt(v[-npca:]))
    if cmpct < 1e-6:
        cmpct = 10.0
        print('Compactness override: %f'%cmpct)

    # Project, redimension as an image with "npca" channels, and segment
    x_pca_subset = (x[use,:] - mu) @ d[:, -npca:]
    del x, mu, d
    x_pca = np.zeros((nc,ns,npca))
    x_pca[use.reshape(nc,ns),:] = x_pca_subset
    del x_pca_subset
    
    x_pca = x_pca.reshape([nc, ns, npca])
    seg_in_chunk = int(sum(use) / float(segsize))

    logging.debug(f'{lstart}: starting slic')
    labels = slic(x_pca, n_segments=seg_in_chunk, compactness=cmpct,
                  max_iter=10, sigma=0, multichannel=True,
                  enforce_connectivity=True, min_size_factor=0.5,
                  max_size_factor=3, mask=use.reshape(nc,ns))

    # Reindex the subscene labels and place them into the larger scene
    labels = labels.reshape([nc * ns])
    labels[np.logical_not(use)] = 0
    labels = labels.reshape([nc, ns])

    logging.info(f'{lstart}: completing')
    return lstart, lend, labels


def segment(spectra: tuple, nodata_value: float, npca: int, segsize: int, nchunk: int, n_cores: int = 1,
            ray_address: str = None, ray_redis_password: str = None,
            ray_temp_dir=None, ray_ip_head=None, logfile=None, loglevel='INFO'):
    """
    Segment an image using SLIC on a PCA.

    Args:
        spectra: tuple of filepaths of image to segment and (optionally) output label file
        nodata_value: data to ignore in radiance image
        npca: number of pca components to use
        segsize: mean segmentation size
        nchunk: size of each image chunk
        n_cores: number of cores to use
        ray_address: ray address to connect to (for multinode implementation)
        ray_redis_password: ray password to use (for multinode implementation)
        ray_temp_dir: ray temp directory to reference
        ray_ip_head: ray ip head to reference (for multinode use)
        logfile: logging file to output to
        loglevel: logging level to use

    """

    logging.basicConfig(format='%(levelname)s:%(message)s', level=loglevel, filename=logfile)

    in_file = spectra[0]
    if len(spectra) > 1 and type(spectra) is tuple:
        lbl_file = spectra[1]
    else:
        lbl_file = spectra + '_lbl'

    # Open input data, get dimensions
    in_img = envi.open(in_file+'.hdr', in_file)
    meta = in_img.metadata
    nl, nb, ns = [int(meta[n]) for n in ('lines', 'bands', 'samples')]

    # Start up a ray instance for parallel work
    rayargs = {'ignore_reinit_error': True,
               'local_mode': n_cores == 1,
               "address": ray_address,
               "_redis_password": ray_redis_password}

    if rayargs['local_mode']:
        rayargs['_temp_dir'] = ray_temp_dir
        # Used to run on a VPN
        ray.services.get_node_ip_address = lambda: '127.0.0.1'

    # We can only set the num_cpus if running on a single-node
    if ray_ip_head is None and ray_redis_password is None:
        rayargs['num_cpus'] = n_cores

    ray.init(**rayargs)
    atexit.register(ray.shutdown)


    # Iterate through image "chunks," segmenting as we go
    all_labels = np.zeros((nl, ns),dtype=np.int64)
    jobs = []
    for lstart in np.arange(0, nl, nchunk):
        # Extract data
        lend = min(lstart+nchunk, nl)

        jobs.append(segment_chunk.remote(lstart, lend, in_file, nodata_value, npca, segsize, logfile=logfile, loglevel=loglevel))

    # Collect results, making sure each chunk is distinct, and enforce an order
    next_label = 1
    rreturn = [ray.get(jid) for jid in jobs]
    for lstart, lend, ret in rreturn:
        if ret is not None:
            logging.debug(f'Collecting chunk: {lstart}')
            chunk_label = ret.copy()
            unique_chunk_labels = np.unique(chunk_label[chunk_label != 0])
            ordered_chunk_labels = np.zeros(chunk_label.shape)
            for lbl in unique_chunk_labels:
                ordered_chunk_labels[chunk_label == lbl] = next_label
                next_label += 1
            all_labels[lstart:lend,...] = ordered_chunk_labels
    del rreturn
    ray.shutdown()

    # Final file I/O
    logging.debug('Writing output')
    lbl_meta = {"samples": str(ns), "lines": str(nl), "bands": "1",
                "header offset": "0", "file type": "ENVI Standard",
                "data type": "4", "interleave": "bil"}
    lbl_img = envi.create_image(lbl_file+'.hdr', lbl_meta, ext='', force=True)
    lbl_mm = lbl_img.open_memmap(interleave='source', writable=True)
    lbl_mm[:, :] = np.array(all_labels, dtype=np.float32).reshape((nl, 1, ns))
    del lbl_mm



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: Philip G. Brodrick, philip.brodrick@jpl.nasa.gov

from collections import OrderedDict
from typing import Dict, List, Type
from isofit.configs.base_config import BaseConfigSection


class InversionConfig(BaseConfigSection):
    """
    Inversion configuration.
    """

    def __init__(self, sub_configdic: dict = None):
        self._windows_type = list()
        self.windows = None
        """List[List[float]]: inversion retrieval windows to operate over."""

        self._cressie_map_confidence_type = bool
        self.cressie_map_confidence = False
        """bool: N. Cressie [ASA 2018] suggests an alternate definition of S_hat for
        more statistically-consistent posterior confidence estimation, this flag runs in this mode"""

        self._mcmc_type = McmcConfig
        self.mcmc = McmcConfig({})
        """MCMC parameters, only used if mode = mcmc."""

        self._integration_grid_type = OrderedDict
        self.integration_grid = OrderedDict({})
        """Grid of inversion points to execute if mode='grid'.  Either fixed, or starting points, depending
        on self.fixed_inversion_grid"""

        self._inversion_grid_as_preseed_type = bool
        self.inversion_grid_as_preseed = False
        """Parameter indicating whether to treat the inversion grid as:
         (True) - a series of seeds for the optimization (variable by the optimization algorithm).
         (False) - a set of fixed points (not variable by the optimization algorithm)
        """

        self._least_squares_params_type = LeastSquaresConfig
        self.least_squares_params = LeastSquaresConfig({})
        """
        Least squares parameters for core inversion solve.  
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html
        for details.
        """

        self.set_config_options(sub_configdic)

    def _check_config_validity(self) -> List[str]:
        errors = list()

        # TODO: add some checking to windows
        if self.windows is None and self.mode != 'simulation':
            errors.append('windows is a required parameters inside of inversion')
        else:
            for subset in self.windows:
                if isinstance(subset, List) is False:
                    errors.append('windows parameter must be a list of lists of wavelength ranges')
                elif subset[0] > subset[1]:
                    errors.append('In inversion window subset {}, wavelength ranges must be in order'.format(subset))
        
        return errors


class McmcConfig(BaseConfigSection):
    """
    MCMC inversion configuration.
    """

    def __init__(self, sub_configdic: dict = None):
        self._iterations_type = int
        self.iterations = 10000
        """int: Number of MCMC iterations to run."""

        self._burnin_type = int
        self.burnin = 200

        self._regularizer_type = float
        self.regularizer = 1e-3

        self._proposal_scaling_type = float
        self.proposal_scaling = 0.01

        self._verbose_type = bool
        self.verbose = True

        self._restart_every_type = int
        self.restart_every = 2000

        self.set_config_options(sub_configdic)

    def _check_config_validity(self) -> List[str]:
        errors = list()

        # TODO: add flags for rile overright, and make sure files don't exist if not checked?

        return errors


class LeastSquaresConfig(BaseConfigSection):
    """
    Least squares config parameters.
    """

    def __init__(self, sub_configdic: dict = None):
        self._method_type = str
        self.method = 'trf'
        """str: Optimzation method to use. Default 'trf'."""

        self._max_nfev_type = int
        self.max_nfev = 20
        """int: Maximum number of function evaluations before the termination. 
        If None (default), the value is chosen automatically. Default 20."""

        self._xtol_type = float
        self.xtol = None
        """float: Tolerance for termination by the change of the independent variables.  
        Default is None, which disables termination from this criteria."""

        self._ftol_type = float
        self.ftol = 0.01
        """float: Tolerance for termination by the change of the cost function.
        Default is 0.01"""

        self._gtol_type = float
        self.gtol = None
        """float: Tolerance for termination by the norm of the gradient. 
        Default is None, which disables termination from this criteria."""

        self._tr_solver_type = str
        self.tr_solver = 'lsmr'
        """str: Method for solving trust-region subproblems, relevant only for ‘trf’ and ‘dogbox’ methods.
        Options are None, 'exact', or 'lsmr'"""

        self.set_config_options(sub_configdic)

    def _check_config_validity(self) -> List[str]:
        errors = list()

        # TODO: add flags for rile overright, and make sure files don't exist if not checked?
        if self.tr_solver is not None:
            if self.tr_solver not in [None, 'exact', 'lsmr']:
                errors.append('Least squares tr_solver must be in [None, \'exact\', \'lsmr\']')

        return errors





#
#  Copyright 2019 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: Philip G Brodrick, philip.brodrick@jpl.nasa.gov
#

import os
import logging
import datetime
from sys import platform
import numpy as np
from copy import deepcopy
import yaml
import pickle
from collections import OrderedDict

from isofit.core.common import resample_spectrum, load_wavelen, VectorInterpolator
from .look_up_tables import TabularRT
from isofit.configs import Config
from isofit.configs.sections.radiative_transfer_config import RadiativeTransferEngineConfig
from isofit.core.sunposition import sunpos
from isofit.radiative_transfer.six_s import SixSRT

from tensorflow import keras
from scipy import interpolate

class SimulatedModtranRT(TabularRT):
    """A hybrid surrogate-model and emulator of MODTRAN-like results

    Args:
        engine_config: the configuration for this particular engine
        full_config (Config): the global configuration
    """   

    def __init__(self, engine_config: RadiativeTransferEngineConfig, full_config: Config):

        # Specify which of the potential MODTRAN LUT parameters are angular, which will be handled differently
        self.angular_lut_keys_degrees = ['OBSZEN', 'TRUEAZ', 'viewzen', 'viewaz', 'solzen', 'solaz']
        self.angular_lut_keys_radians = []

        super().__init__(engine_config, full_config)

        self.lut_quantities = ['rhoatm', 'transm', 'sphalb', 'transup']
        self.treat_as_emissive = False

        # Load in the emulator aux - hold off on emulator till last
        # second, as the model is large, and we don't want to load in 
        # parallel if possible
        emulator_aux = np.load(engine_config.emulator_aux_file)

        simulator_wavelengths = emulator_aux['simulator_wavelengths']
        simulator_wavelengths = np.arange(350, 2500+2.5, 2.5)
        emulator_wavelengths = emulator_aux['emulator_wavelengths']
        n_simulator_chan = len(simulator_wavelengths)
        n_emulator_chan = len(emulator_wavelengths)

        for lq in self.lut_quantities:
            if lq not in emulator_aux['rt_quantities'].tolist() and lq != 'transup':
                raise AttributeError('lut_quantities: {} do not match emulator_aux rt_quantities: {}'.format(self.lut_quantities, emulator_aux['rt_quantities']))

        interpolator_disk_paths = [engine_config.interpolator_base_path + '_' + rtq + '.pkl' for rtq in self.lut_quantities]

        # Build a new config for sixs simulation runs using existing config
        sixs_config: RadiativeTransferEngineConfig = deepcopy(engine_config)
        sixs_config.aerosol_model_file = None
        sixs_config.aerosol_template_file = None
        sixs_config.emulator_file = None

        # Populate the sixs-values from the modtran template file
        with open(sixs_config.template_file, 'r') as infile:
            modtran_input = yaml.safe_load(infile)['MODTRAN'][0]['MODTRANINPUT']

        # Do a quickk conversion to put things in solar azimuth/zenith terms for 6s
        dt = datetime.datetime(2000, 1, 1) + datetime.timedelta(days=modtran_input['GEOMETRY']['IDAY'] - 1)  + \
             datetime.timedelta(hours = modtran_input['GEOMETRY']['GMTIME'])

        solar_azimuth, solar_zenith, ra, dec, h = sunpos(dt, modtran_input['GEOMETRY']['PARM1'],-modtran_input['GEOMETRY']['PARM2'],modtran_input['SURFACE']['GNDALT'] * 1000.0)  

        sixs_config.day = dt.day
        sixs_config.month = dt.month
        sixs_config.elev = modtran_input['SURFACE']['GNDALT']
        sixs_config.alt = modtran_input['GEOMETRY']['H1ALT']
        sixs_config.solzen = solar_zenith
        sixs_config.solaz = solar_azimuth
        sixs_config.viewzen = 180 - modtran_input['GEOMETRY']['OBSZEN']
        sixs_config.viewaz = modtran_input['GEOMETRY']['TRUEAZ']
        sixs_config.wlinf = 0.35
        sixs_config.wlsup = 2.5



        # Build the simulator
        logging.debug('Create RTE simulator')
        sixs_rte = SixSRT(sixs_config, full_config, build_lut_only = False, 
                          wavelength_override=simulator_wavelengths, fwhm_override=np.ones(n_simulator_chan)*2.,
                          modtran_emulation=True)
        self.solar_irr = sixs_rte.solar_irr
        self.esd = sixs_rte.esd
        self.coszen = sixs_rte.coszen

        emulator_irr = emulator_aux['solar_irr']
        irr_factor_ref = sixs_rte.esd[200, 1]
        irr_factor_current = sixs_rte.esd[sixs_rte.day_of_year-1, 1]

        # First, convert our irr to date 0, then back to current date
        self.solar_irr = resample_spectrum(emulator_irr  * irr_factor_ref**2 / irr_factor_current**2, emulator_wavelengths, self.wl, self.fwhm)

        # First, check if we've already got the vector interpolators built on disk:
        prebuilt = np.all([os.path.isfile(x) for x in interpolator_disk_paths])
        if prebuilt == False:
            # Load the emulator
            emulator = keras.models.load_model(engine_config.emulator_file)

            # Couple emulator-simulator
            emulator_inputs = np.zeros((sixs_rte.points.shape[0],n_simulator_chan*len(emulator_aux['rt_quantities'])), dtype=float)

            logging.info('loading 6s results for emulator')
            for ind, (point, fn) in enumerate(zip(self.points, sixs_rte.files)):
                simulator_output = sixs_rte.load_rt(fn, resample=False)
                for keyind, key in enumerate(emulator_aux['rt_quantities']):
                    emulator_inputs[ind,keyind*n_simulator_chan:(keyind+1)*n_simulator_chan] = simulator_output[key]
            emulator_inputs[np.isnan(emulator_inputs)] = 0
            emulator_inputs_match_output = np.zeros((emulator_inputs.shape[0],n_emulator_chan*len(emulator_aux['rt_quantities'])))
            for key_ind, key in enumerate(emulator_aux['rt_quantities']):
                band_range_o = np.arange(n_emulator_chan * key_ind, n_emulator_chan * (key_ind + 1))
                band_range_i = np.arange(n_simulator_chan * key_ind, n_simulator_chan * (key_ind + 1))

                finterp = interpolate.interp1d(simulator_wavelengths,emulator_inputs[:,band_range_i])
                emulator_inputs_match_output[:,band_range_o] = finterp(emulator_wavelengths)
 
            if 'response_scaler' in emulator_aux.keys():
                response_scaler = emulator_aux['response_scaler']
            else:
                response_scaler = 100.

            logging.debug('Emulating')
            emulator_outputs = emulator.predict(emulator_inputs) / response_scaler
            emulator_outputs = emulator_outputs + emulator_inputs_match_output

            dims_aug = self.lut_dims + [self.n_chan]
            for key_ind, key in enumerate(emulator_aux['rt_quantities']):
                interpolator_inputs = np.zeros(dims_aug, dtype=float)
                for point_ind, point in enumerate(self.points):
                    ind = [np.where(g == p)[0] for g, p in
                           zip(self.lut_grids, point)]
                    ind = tuple(ind)
                    interpolator_inputs[ind] = resample_spectrum(emulator_outputs[point_ind, key_ind*n_emulator_chan: (key_ind+1)*n_emulator_chan], emulator_wavelengths, self.wl, self.fwhm)

                if key == 'transup':
                    interpolator_inputs[...] = 0

                self.luts[key] = VectorInterpolator(self.lut_grids, interpolator_inputs,
                                                    self.lut_interp_types)

            self.luts['transup'] = VectorInterpolator(self.lut_grids, interpolator_inputs*0, self.lut_interp_types )
            
            for key_ind, key in enumerate(self.lut_quantities):
                with open(interpolator_disk_paths[key_ind], 'wb') as fi:
                    pickle.dump(self.luts[key], fi, protocol=4)

        else:
            logging.info('Prebuilt LUT interpolators found, loading from disk')
            for key_ind, key in enumerate(self.lut_quantities):
                with open(interpolator_disk_paths[key_ind], 'rb') as fi:
                    self.luts[key] = pickle.load(fi)


        
    def recursive_dict_search(self, indict, key):
        for k, v  in indict.items():
            if k == key:
                return v
            elif isinstance(v, dict):
                found = self.recursive_dict_search(v, key) 
                if found is not None: 
                    return found 


    def _lookup_lut(self, point):
        ret = {}
        for key, lut in self.luts.items():
            ret[key] = np.array(lut(point)).ravel()
        return ret


    def get(self, x_RT, geom):
        point = np.zeros((self.n_point,))
        for point_ind, name in enumerate(self.lut_grid_config):
            if name in self.statevector_names:
                ix = self.statevector_names.index(name)
                point[point_ind] = x_RT[ix]
            elif name == "OBSZEN":
                point[point_ind] = geom.OBSZEN
            elif name == "GNDALT":
                point[point_ind] = geom.surface_elevation_km
            elif name == "H1ALT":
                point[point_ind] = geom.observer_altitude_km
            elif name == "viewzen":
                point[point_ind] = geom.observer_zenith
            elif name == "viewaz":
                point[point_ind] = geom.observer_azimuth
            elif name == "solaz":
                point[point_ind] = geom.solar_azimuth
            elif name == "solzen":
                point[point_ind] = geom.solar_zenith
            elif name == "TRUEAZ":
                point[point_ind] = geom.TRUEAZ
            elif name == 'phi':
                point[point_ind] = geom.phi
            elif name == 'umu':
                point[point_ind] = geom.umu
            else:
                # If a variable is defined in the lookup table but not
                # specified elsewhere, we will default to the minimum
                point[point_ind] = min(self.lut_grid_config[name])

        return self._lookup_lut(point)

    def get_L_atm(self, x_RT, geom):
        r = self.get(x_RT, geom)
        rho = r['rhoatm']
        rdn = rho / np.pi*(self.solar_irr * self.coszen)
        return rdn

    def get_L_down_transmitted(self, x_RT, geom):
        r = self.get(x_RT, geom)
        rdn = (self.solar_irr * self.coszen) / np.pi * r['transm']
        return rdn



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import numpy as np
from scipy.interpolate import interp1d

from ..core.common import load_spectrum, load_wavelen
from isofit.configs import Config


class Surface:
    """A model of the surface.

    Surface models are stored as MATLAB '.mat' format files.
    """

    def __init__(self, full_config: Config):

        config = full_config.forward_model.surface

        self.statevec_names = []
        self.bounds = np.array([])
        self.scale = np.array([])
        self.init = np.array([])
        self.bvec = []
        self.bval = np.array([])
        self.emissive = False

        if config.reflectance is not None:
            self.rfl = config['reflectance']
            self.resample_reflectance()
        if config.reflectance_file is not None:
            self.rfl, self.rwl = load_spectrum(config.reflectance_file)
            self.wl = self.rwl.copy()
            self.n_wl = len(self.wl)
        if config.wavelength_file is not None:
            self.wl, self.fwhm = load_wavelen(config.wavelength_file)
            self.n_wl = len(self.wl)
            self.resample_reflectance()

    def resample_reflectance(self):
        """Make sure model wavelengths align with the wavelength file."""

        if hasattr(self, 'rwl'):
            p = interp1d(self.rwl, self.rfl, fill_value='extrapolate')
            self.rfl = p(self.wl)

    def xa(self, x_surface, geom):
        """Mean of prior state vector distribution calculated at state x."""

        return np.array(self.init)

    def Sa(self, x_surface, geom):
        """Covariance of prior state vector distribution calculated at state x."""

        return np.zeros((0, 0), dtype=float)

    def fit_params(self, rfl_meas, geom, *args):
        """Given a directional reflectance estimate and one or more emissive 
        parameters, fit a state vector."""

        return np.array([])

    def calc_lamb(self, x_surface, geom):
        """Calculate a Lambertian surface reflectance for this state vector."""

        return self.rfl

    def calc_rfl(self, x_surface, geom):
        """Calculate the directed reflectance (specifically the HRDF) for this
        state vector."""

        return self.rfl

    def drfl_dsurface(self, x_surface, geom):
        """Partial derivative of reflectance with respect to state vector, 
        calculated at x_surface. In the case that there are no free 
        paramters our convention is to return the vector of zeros."""

        return np.zeros((self.n_wl, 1))

    def drfl_dsurfaceb(self, x_surface, geom):
        """Partial derivative of reflectance with respect to unmodeled 
        variables, calculated at x_surface. In the case that there are no
        free paramters our convention is to return the vector of zeros."""

        return np.zeros((self.n_wl, 1))

    def calc_Ls(self, x_surface, geom):
        """Emission of surface, as a radiance."""

        return np.zeros((self.n_wl,))

    def dLs_dsurface(self, x_surface, geom):
        """Partial derivative of surface emission with respect to state vector, 
        calculated at x_surface. In the case that there are no free paramters
        our convention is to return the vector of zeros."""

        return np.zeros((self.n_wl, 1))

    def summarize(self, x_surface, geom):
        """Summary of state vector."""

        return ''



import numpy as np
from isofit.core.common import max_table_size, eps, combos


def test_max_table_size():
    assert max_table_size == 500


def test_eps():
    assert eps == 1e-5


def test_combos():
    inds = np.array([[1, 2], [3, 4, 5]])
    result = np.array([[1, 3], [2, 3], [1, 4], [2, 4], [1, 5], [2, 5]])
    assert np.array_equal(combos(inds), result)



import scipy as s
from isofit.core.fileio import typemap, max_frames_size, flush_rate


def test_typemap():
    assert typemap[s.uint64] == 15


def test_max_frames_size():
    assert max_frames_size == 100


def test_flush_rate():
    assert flush_rate == 10



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: Philip G. Brodrick, philip.brodrick@jpl.nasa.gov

from typing import Dict, List, Type
from isofit.configs.base_config import BaseConfigSection
from isofit.configs.sections.statevector_config import StateVectorConfig
import os
from collections import OrderedDict
import logging


class RadiativeTransferEngineConfig(BaseConfigSection):
    """
    Radiative transfer unknowns configuration.
    """

    def __init__(self, sub_configdic: dict = None, name: str = None):
        self._name_type = str
        self.name = name
        """str: Name of config - optional, and not currently used."""

        self._engine_name_type = str
        self.engine_name = None
        """str: Name of radiative transfer engine to use - options ['modtran', 'libradtran', '6s']."""

        self._engine_base_dir_type = str
        self.engine_base_dir = None
        """str: base directory of the given radiative transfer engine on user's OS."""

        self._wavelength_range_type = list()
        self.wavelength_range = None
        """List: The wavelength range to execute this radiative transfer engine over."""

        self._environment_type = str
        self.environment = None
        """str: Additional environment directives for the shell script."""

        self._lut_path_type = str
        self.lut_path = None
        """str: The path to the look up table directory used by the radiative transfer engine."""

        self._template_file_type = str
        self.template_file = None
        """str: A template file to be used as the base-configuration for the given radiative transfer engine."""

        self._lut_names_type = list()
        self.lut_names = None
        """List: Names of the elements to run this radiative transfer element on.  Must be a subset
        of the keys in radiative_transfer->lut_grid.  If not specified, uses all keys from 
        radiative_transfer-> lut_grid.  Auto-sorted (alphabetically) below."""

        self._statevector_names_type = list()
        self.statevector_names = None
        """List: Names of the statevector elements to use with this radiative transfer engine.  Must be a subset
        of the keys in radiative_transfer->statevector.  If not specified, uses all keys from 
        radiative_transfer->statevector.  Auto-sorted (alphabetically) below."""

        # MODTRAN parameters
        self._aerosol_template_file_type = str
        self.aerosol_template_file = None
        """str: Aerosol template file, currently only implemented for MODTRAN."""

        self._aerosol_model_file_type = str
        self.aerosol_model_file = None
        """str: Aerosol model file, currently only implemented for MODTRAN."""

        self._multipart_transmittance_type = bool
        self.multipart_transmittance = False
        """str: Use True to specify triple-run diffuse & direct transmittance 
           estimation.  Only implemented for MODTRAN."""

        # MODTRAN simulator
        self._emulator_file_type = str
        self.emulator_file = None
        """str: Path to emulator model file"""

        self._emulator_aux_file_type = str
        self.emulator_aux_file = None
        """str: path to emulator auxiliary data - expected npz format"""

        self._interpolator_base_path_type = str
        self.interpolator_base_path = None
        """str: path to emulator interpolator base - will dump multiple pkl extensions to this location"""

        # 6S parameters - not the corcommemnd
        # TODO: these should come from a template file, as in modtran
        self._day_type = int
        self.day = None
        """int: 6s-only day parameter."""

        self._month_type = int
        self.month = None
        """int: 6s-only month parameter."""

        self._elev_type = float
        self.elev = None
        """float: 6s-only elevation parameter."""

        self._alt_type = float
        self.alt = None
        """float: 6s-only altitude parameter."""

        self._obs_file_type = str
        self.obs_file = None
        """str: 6s-only observation file."""

        self._solzen_type = float
        self.solzen = None
        """float: 6s-only solar zenith."""

        self._solaz_type = float
        self.solaz = None
        """float: 6s-only solar azimuth."""

        self._viewzen_type = float
        self.viewzen = None
        """float: 6s-only view zenith."""

        self._viewaz_type = float
        self.viewaz = None
        """float: 6s-only view azimuth."""

        self._earth_sun_distance_file_type = str
        self.earth_sun_distance_file = None
        """str: 6s-only earth-to-sun distance file."""

        self._irradiance_file_type = str
        self.irradiance_file = None
        """str: 6s-only irradiance file."""

        self.set_config_options(sub_configdic)

        if self.lut_names is not None:
            self.lut_names.sort()

        if self.statevector_names is not None:
            self.statevector_names.sort()
        
        if self.interpolator_base_path is None and self.emulator_file is not None:
            self.interpolator_base_path = self.emulator_file + '_interpolator'
            logging.info('No interpolator base path set, and emulator used, so auto-setting interpolator path at: {}'.format(self.interpolator_base_path))


    def _check_config_validity(self) -> List[str]:
        errors = list()

        # Check that all input files exist
        for key in self._get_nontype_attributes():
            value = getattr(self, key)
            if value is not None and key[-5:] == '_file' and key != 'emulator_file':
                if os.path.isfile(value) is False:
                    errors.append('Config value radiative_transfer->{}: {} not found'.format(key, value))

        valid_rt_engines = ['modtran', 'libradtran', '6s', 'simulated_modtran']
        if self.engine_name not in valid_rt_engines:
            errors.append('radiative_transfer->raditive_transfer_model: {} not in one of the available models: {}'.
                          format(self.engine_name, valid_rt_engines))

        if self.multipart_trasmittance and self.engine_name != 'modtran':
            errors.append('Multipart transmittance is supported for MODTRAN only')

        if self.earth_sun_distance_file is None and self.engine_name == '6s':
            errors.append('6s requires earth_sun_distance_file to be specified')

        if self.irradiance_file is None and self.engine_name == '6s':
            errors.append('6s requires irradiance_file to be specified')

        if self.engine_name == 'simulated_modtran' and self.emulator_file is None:
            errors.append('The Modtran Simulator requires an emulator_file to be specified.')

        if self.engine_name == 'simulated_modtran' and self.emulator_aux_file is None:
            errors.append('The Modtran Simulator requires an emulator_aux_file to be specified.')

        files = [self.earth_sun_distance_file, self.irradiance_file,
                 self.obs_file, self.aerosol_model_file, self.aerosol_template_file]
        for f in files:
            if f is not None:
                if os.path.isfile(f) is False:
                    errors.append('Radiative transfer engine file not found on system: {}'.
                                  format(self.earth_sun_distance_file))
        return errors


class RadiativeTransferUnknownsConfig(BaseConfigSection):
    """
    Radiative transfer unknowns configuration.
    """

    def __init__(self, sub_configdic: dict = None):
        self._H2O_ABSCO_type = float
        self.H2O_ABSCO = None

        self.set_config_options(sub_configdic)

    def _check_config_validity(self) -> List[str]:
        errors = list()

        return errors


class RadiativeTransferConfig(BaseConfigSection):
    """
    Forward model configuration.
    """

    def __init__(self, sub_configdic: dict = None):

        self._statevector_type = StateVectorConfig
        self.statevector: StateVectorConfig = StateVectorConfig({})

        self._lut_grid_type = OrderedDict
        self.lut_grid = None

        self._unknowns_type = RadiativeTransferUnknownsConfig
        self.unknowns: RadiativeTransferUnknownsConfig = None

        self.set_config_options(sub_configdic)

        # sort lut_grid
        for key, value in self.lut_grid.items():
            self.lut_grid[key] = sorted(self.lut_grid[key])
        self.lut_grid = OrderedDict(sorted(self.lut_grid.items(), key=lambda t: t[0]))

        # Hold this parameter for after the config_options, as radiative_transfer_engines
        # have a special (dynamic) load
        self._radiative_transfer_engines_type = list()
        self.radiative_transfer_engines = []

        self._set_rt_config_options(sub_configdic['radiative_transfer_engines'])

    def _set_rt_config_options(self, subconfig):
        if type(subconfig) is list:
            for rte in subconfig:
                rt_model = RadiativeTransferEngineConfig(rte)
                self.radiative_transfer_engines.append(rt_model)
        elif type(subconfig) is dict:
            for key in subconfig:
                rt_model = RadiativeTransferEngineConfig(subconfig[key], name=key)
                self.radiative_transfer_engines.append(rt_model)

    def _check_config_validity(self) -> List[str]:
        errors = list()

        for key, item in self.lut_grid.items():
            if len(item) < 2:
                errors.append('lut_grid item {} has less than the required 2 elements'.format(key))
        
        for rte in self.radiative_transfer_engines:
            errors.extend(rte.check_config_validity())

        return errors



from .surface import Surface
from .surface_multicomp import MultiComponentSurface
from .surface_glint import GlintSurface
from .surface_thermal import ThermalSurface
from .surface_thermal import ThermalSurface
from .surface_lut import LUTSurface



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: Philip G. Brodrick, philip.brodrick@jpl.nasa.gov

from collections import OrderedDict
from typing import Dict, List, Type
import numpy as np
import logging


class BaseConfigSection(object):
    """
    Base Configuration Section from which all Configuration Sections inherit. Handles shared functionality like getting,
    setting, and cleaning configuration options.
    """

    def __init__(self) -> None:
        return

    def set_config_options(self, configdict: dict = None) -> None:
        """ Read dictionary and assign to attributes, leaning on _set_callable_attributes
        Args:
            configdict: dictionary-style config for parsing
        """
        for key in self._get_nontype_attributes():
            keytype = getattr(self, '_' + key + '_type')
            if key in configdict:
                if callable(keytype):
                    sub_config = keytype(configdict[key])
                    setattr(self, key, sub_config)
                else:
                    setattr(self, key, configdict[key])
        return

    def check_config_validity(self) -> List[str]:
        errors = list()
        message_type = (
            "Invalid type for config option {} in config section {}. The provided value {} is a {}, "
            + "but the required value should be a {}."
        )

        # First check typing
        for key in self._get_nontype_attributes():

            # get the actual parameter value
            value = getattr(self, key)

            # None assignments are fine
            if value is None:
                continue

            # check it against expected
            type_expected = self._get_expected_type_for_option_key(key)
            # Lists are complicated, retype
            if isinstance(type_expected, List):
                type_expected = List

            if isinstance(value, type_expected):
                continue

            # At this point, we have a type mismatch, add to error list
            errors.append(message_type.format(key, self.__class__.__name__,
                                              value, type(value), type_expected))

        errors.extend(self._check_config_validity())

        # Now do a full check on each submodule
        for key in self._get_nontype_attributes():
            value = getattr(self, key)
            try:
                logging.debug('Configuration check of: {}'.format(key))
                errors.extend(value.check_config_validity())
            except AttributeError:
                logging.debug('Configuration check: {} is not an object, skipping'.format(key))

        return errors

    def get_config_options_as_dict(self) -> Dict[str, Dict[str, any]]:
        config_options = OrderedDict()
        for key in self._get_nontype_attributes():
            value = getattr(self, key)
            if type(value) is tuple:
                value = list(value)  # Lists look nicer in config files and seem friendlier
            config_options[key] = value
        return config_options

    def _check_config_validity(self) -> List[str]:
        return list()

    def _get_expected_type_for_option_key(self, option_key: str) -> type:
        return getattr(self, "_{}_type".format(option_key))

    def _get_nontype_attributes(self) -> List[str]:
        keys = []
        for key in self.__dict__.keys():
            if key[0] == '_':
                continue
            keys.append(key)
        return keys

    def _get_type_attributes(self) -> List[str]:
        keys = []
        for key in self.__dict__.keys():
            if key[0] == '_' and key[-5:] == '_type':
                keys.append(key)
        return keys

    def _get_hidden_attributes(self) -> List[str]:
        keys = []
        for key in self.__dict__.keys():
            if key[0] == '_' and key[-5:] != '_type':
                keys.append(key)
        return keys

    def get_all_elements(self):
        return [getattr(self, x) for x in self._get_nontype_attributes()]

    def get_all_element_names(self):
        return self._get_nontype_attributes()

    def get_elements(self):
        elements = self.get_all_elements()
        element_names = self._get_nontype_attributes()
        valid = [x is not None for x in elements]
        elements = [elements[x] for x in range(len(elements)) if valid[x]]
        element_names = [element_names[x] for x in range(len(element_names)) if valid[x]]

        order = np.argsort(element_names)
        elements = [elements[idx] for idx in order]
        element_names = [element_names[idx] for idx in order]

        return elements, element_names

    def get_element_names(self):
        elements, element_names = self.get_elements()
        return element_names

    def get_single_element_by_name(self, name):
        elements, element_names = self.get_elements()
        return elements[element_names.index(name)]





#from .input_config import InputConfig
#from .output_config import OutputConfig
#from .forward_model_config import ForwardModelConfig
#from .instrument_config import InstrumentConfig
#from .radiative_transfer_config import RadiativeTransferModelConfig
#from .radiative_transfer_config import RadiativeTransferConfig
#from .radiative_transfer_config import RadiativeTransferUnknownsConfig



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: Philip G. Brodrick, philip.brodrick@jpl.nasa.gov

from typing import Dict, List, Type
from isofit.configs.base_config import BaseConfigSection
from isofit.configs.sections.statevector_config import StateVectorConfig
import os


class InstrumentUnknowns(BaseConfigSection):
    """
    Instrument Unknowns configuration.
    """

    def __init__(self, sub_configdic: dict = None):
        self._channelized_radiometric_uncertainty_file_type = str
        self.channelized_radiometric_uncertainty_file = None

        self._uncorrelated_radiometric_uncertainty_type = float
        self.uncorrelated_radiometric_uncertainty = None

        self._wavelength_calibration_uncertainty_type = float
        self.wavelength_calibration_uncertainty = None

        self._stray_srf_uncertainty_type = float
        self.stray_srf_uncertainty = None

        self.set_config_options(sub_configdic)

    def _check_config_validity(self) -> List[str]:
        errors = list()

        file_params = [self.channelized_radiometric_uncertainty_file]
        for param in file_params:
            if param is not None:
                if os.path.isfile(param) is False:
                    errors.append('Instrument unknown file: {} not found'.format(param))

        return errors


class InstrumentConfig(BaseConfigSection):
    """
    Instrument configuration.
    """

    def __init__(self, sub_configdic: dict = None):

        self._wavelength_file_type = str
        self.wavelength_file = None

        self._integrations_type = int
        self.integrations = None
        """Number of integrations comprising the measurement.  
        Noise diminishes with the square root of this number.  Applicable in concert with parametric_noise_file 
        or pushbroom_noise_file"""

        self._unknowns_type = InstrumentUnknowns
        self.unknowns: InstrumentUnknowns = None

        self._fast_resample_type = bool
        self.fast_resample = True
        """bool: Approximates a complete resampling by a convolution with a uniform FWHM."""

        self._statevector_type = StateVectorConfig
        self.statevector: StateVectorConfig = None

        self._SNR_type = float
        self.SNR = None
        """float: We have several ways to define the instrument noise.  The simplest model is based on a single uniform 
        SNR number that is signal-independnet and applied uniformly to all wavelengths"""

        self._parametric_noise_file_type = str
        self.parametric_noise_file = None
        """str: We have several ways to define the instrument noise.
        The second option is a parametric, signal- and wavelength-
        dependent noise function. This is given by a four-column
        ASCII Text file.  Rows represent, respectively, the reference
        wavelength, and coefficients A, B, and C that define the
        noise-equivalent radiance via NeDL = A * sqrt(B+L) + C
        For the actual radiance L."""

        self._pushbroom_noise_file_type = str
        self.pushbroom_noise_file = None
        """str: We have several ways to define the instrument noise.
        The third option is a full pushbroom noise model that
        specifies noise columns and covariances independently for
        each cross-track location via an ENVI-format binary data file."""

        self._nedt_noise_file_type = str
        self.nedt_noise_file = None
        """str: We have several ways to define the instrument noise.  The last is NEDT noise"""

        self.set_config_options(sub_configdic)

        # If necessary, initialize some blank options
        if self.statevector is None:
            self.statevector = StateVectorConfig({})

    def _check_config_validity(self) -> List[str]:
        errors = list()

        noise_options = [self.SNR, self.parametric_noise_file,
                         self.pushbroom_noise_file, self.nedt_noise_file]
        used_noise_options = [x for x in noise_options if x is not None]

        if len(used_noise_options) == 0:
            errors.append('Instrument noise not defined.')

        if len(used_noise_options) > 1:
            errors.append('Multiple instrument noise options selected - please choose only 1.')

        file_params = [self.parametric_noise_file, self.pushbroom_noise_file, self.nedt_noise_file]
        for param in file_params:
            if param is not None:
                if os.path.isfile(param) is False:
                    errors.append('Instrument config file: {} not found'.format(param))

        return errors



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import os
import numpy as np
import scipy.io
import pylab as plt
from spectral.io import envi
import logging
from collections import OrderedDict

from .common import load_spectrum, eps, resample_spectrum
from isofit.inversion.inverse_simple import invert_simple, invert_algebraic
from .geometry import Geometry
from isofit.configs import Config
from isofit.core.forward import ForwardModel


### Variables ###

# Constants related to file I/O
typemap = {
    np.uint8: 1,
    np.int16: 2,
    np.int32: 3,
    np.float32: 4,
    np.float64: 5,
    np.complex64: 6,
    np.complex128: 9,
    np.uint16: 12,
    np.uint32: 13,
    np.int64: 14,
    np.uint64: 15
}

max_frames_size = 100
flush_rate = 10


### Classes ###

class SpectrumFile:
    """A buffered file object that contains configuration information about formatting, etc."""

    def __init__(self, fname, write=False, n_rows=None, n_cols=None, n_bands=None,
                 interleave=None, dtype=np.float32, wavelengths=None, fwhm=None,
                 band_names=None, bad_bands='[]', zrange='{0.0, 1.0}', flag=-9999.0,
                 ztitles='{Wavelength (nm), Magnitude}', map_info='{}'):
        """."""

        self.frames = OrderedDict()
        self.write = write
        self.fname = os.path.abspath(fname)
        self.wl = wavelengths
        self.band_names = band_names
        self.fwhm = fwhm
        self.flag = flag
        self.n_rows = n_rows
        self.n_cols = n_cols
        self.n_bands = n_bands

        if self.fname.endswith('.txt'):

            # The .txt suffix implies a space-separated ASCII text file of
            # one or more data columns.  This is cheap to load and store, so
            # we do not defer read/write operations.
            logging.debug('Inferred ASCII file format for %s' % self.fname)
            self.format = 'ASCII'
            if not self.write:
                self.data,  self.wl = load_spectrum(self.fname)
                self.n_rows, self.n_cols, self.map_info = 1, 1, '{}'
                if self.wl is not None:
                    self.n_bands = len(self.wl)
                else:
                    self.n_bands = None
                self.meta = {}

        elif self.fname.endswith('.mat'):

            # The .mat suffix implies a matlab-style file, i.e. a dictionary
            # of 2D arrays and other matlab-like objects. This is typically
            # only used for specific output products associated with single
            # spectrum retrievals; there is no read option.
            logging.debug('Inferred MATLAB file format for %s' % self.fname)
            self.format = 'MATLAB'
            if not self.write:
                logging.error('Unsupported MATLAB file in input block')
                raise IOError('MATLAB format in input block not supported')

        else:

            # Otherwise we assume it is an ENVI-format file, which is
            # basically just a binary data cube with a detached human-
            # readable ASCII header describing dimensions, interleave, and
            # metadata.  We buffer this data in self.frames, reading and
            # writing individual rows of the cube on-demand.
            logging.debug('Inferred ENVI file format for %s' % self.fname)
            self.format = 'ENVI'

            if not self.write:

                # If we are an input file, the header must preexist.
                if not os.path.exists(self.fname+'.hdr'):
                    logging.error('Could not find %s' % (self.fname+'.hdr'))
                    raise IOError('Could not find %s' % (self.fname+'.hdr'))

                # open file and copy metadata
                self.file = envi.open(self.fname + '.hdr', fname)
                self.meta = self.file.metadata.copy()

                self.n_rows = int(self.meta['lines'])
                self.n_cols = int(self.meta['samples'])
                self.n_bands = int(self.meta['bands'])
                if 'data ignore value' in self.meta:
                    self.flag = float(self.meta['data ignore value'])
                else:
                    self.flag = -9999.0

            else:

                # If we are an output file, we may have to build the header
                # from scratch.  Hopefully the caller has supplied the
                # necessary metadata details.
                meta = {
                    'lines': n_rows,
                    'samples': n_cols,
                    'bands': n_bands,
                    'byte order': 0,
                    'header offset': 0,
                    'map info': map_info,
                    'file_type': 'ENVI Standard',
                    'sensor type': 'unknown',
                    'interleave': interleave,
                    'data type': typemap[dtype],
                    'wavelength units': 'nm',
                    'z plot range': zrange,
                    'z plot titles': ztitles,
                    'fwhm': fwhm,
                    'bbl': bad_bands,
                    'band names': band_names,
                    'wavelength': self.wl
                }

                for k, v in meta.items():
                    if v is None:
                        logging.error('Must specify %s' % (k))
                        raise IOError('Must specify %s' % (k))

                if os.path.isfile(fname+'.hdr') is False:
                    self.file = envi.create_image(fname+'.hdr', meta, ext='',
                                                  force=True)
                else:
                    self.file = envi.open(fname+'.hdr')

            self.open_map_with_retries()

    def open_map_with_retries(self):
        """Try to open a memory map, handling Beowulf I/O issues."""

        self.memmap = None
        for attempt in range(10):
            self.memmap = self.file.open_memmap(interleave='bip',
                                                writable=self.write)
            if self.memmap is not None:
                return
        raise IOError('could not open memmap for '+self.fname)

    def get_frame(self, row):
        """The frame is a 2D array, essentially a list of spectra. The 
        self.frames list acts as a hash table to avoid storing the 
        entire cube in memory. So we read them or create them on an
        as-needed basis.  When the buffer flushes via a call to 
        flush_buffers, they will be deleted."""

        if row not in self.frames:
            if not self.write:
                d = self.memmap[row, :, :]
                self.frames[row] = d.copy()
            else:
                self.frames[row] = np.nan * np.zeros((self.n_cols, self.n_bands))
        return self.frames[row]

    def write_spectrum(self, row, col, x):
        """We write a spectrum. If a binary format file, we simply change
        the data cached in self.frames and defer file I/O until 
        flush_buffers is called."""

        if self.format == 'ASCII':

            # Multicolumn output for ASCII products
            np.savetxt(self.fname, x, fmt='%10.6f')

        elif self.format == 'MATLAB':

            # Dictionary output for MATLAB products
            scipy.io.savemat(self.fname, x)

        else:

            # Omit wavelength column for spectral products
            frame = self.get_frame(row)
            if x.ndim == 2:
                x = x[:, -1]
            frame[col, :] = x

    def read_spectrum(self, row, col):
        """Get a spectrum from the frame list or ASCII file. Note that if
        we are an ASCII file, we have already read the single spectrum and 
        return it as-is (ignoring the row/column indices)."""

        if self.format == 'ASCII':
            return self.data
        else:
            frame = self.get_frame(row)
            return frame[col]

    def flush_buffers(self):
        """Write to file, and refresh the memory map object."""

        if self.format == 'ENVI':
            if self.write:
                for row, frame in self.frames.items():
                    valid = np.logical_not(np.isnan(frame[:, 0]))
                    self.memmap[row, valid, :] = frame[valid, :]
            self.frames = OrderedDict()
            del self.file
            self.file = envi.open(self.fname+'.hdr', self.fname)
            self.open_map_with_retries()


class IO:
    """..."""

    def __init__(self, config: Config, forward: ForwardModel, inverse, active_rows, active_cols):
        """Initialization specifies retrieval subwindows for calculating
        measurement cost distributions."""

        self.input = config.input
        self.output = config.output

        self.iv = inverse
        self.fm = forward
        self.bbl = '[]'
        self.radiance_correction = None
        self.meas_wl = forward.instrument.wl_init
        self.meas_fwhm = forward.instrument.fwhm_init
        self.writes = 0
        self.n_rows = 1
        self.n_cols = 1
        self.n_sv = len(self.fm.statevec)
        self.n_chan = len(self.fm.instrument.wl_init)

        self.simulation_mode = config.implementation.mode == 'simulation'

        # Names of either the wavelength or statevector outputs
        wl_names = [('Channel %i' % i) for i in range(self.n_chan)]
        sv_names = self.fm.statevec.copy()

        self.defined_outputs, self.defined_inputs = {}, {}
        self.infiles, self.outfiles, self.map_info = {}, {}, '{}'

        # Load input files and record relevant metadata
        for element, element_name in zip(*self.input.get_elements()):
            self.infiles[element_name] = SpectrumFile(element)

            if (self.infiles[element_name].n_rows > self.n_rows) or \
               (self.infiles[element_name].n_cols > self.n_cols):
                self.n_rows = self.infiles[element_name].n_rows
                self.n_cols = self.infiles[element_name].n_cols

            for inherit in ['map info', 'bbl']:
                if inherit in self.infiles[element_name].meta:
                    setattr(self, inherit.replace(' ', '_'),
                            self.infiles[element_name].meta[inherit])

        for element, element_header, element_name in zip(*self.output.get_output_files()):
            band_names, ztitle, zrange = element_header

            if band_names == 'statevector':
                band_names = sv_names
            elif band_names == 'wavelength':
                band_names = wl_names
            elif band_names == 'atm_coeffs':
                band_names = wl_names*5
            else:
                band_names = '{}'

            n_bands = len(band_names)
            self.outfiles[element_name] = SpectrumFile(
                element,
                write=True,
                n_rows=self.n_rows,
                n_cols=self.n_cols,
                n_bands=n_bands,
                interleave='bip',
                dtype=np.float32,
                wavelengths=self.meas_wl,
                fwhm=self.meas_fwhm,
                band_names=band_names,
                bad_bands=self.bbl,
                map_info=self.map_info,
                zrange=zrange,
                ztitles=ztitle
            )

        # Do we apply a radiance correction?
        if self.input.radiometry_correction_file is not None:
            filename = self.input.radiometry_correction_file
            self.radiance_correction, wl = load_spectrum(filename)

        # Last thing is to define the active image area
        if active_rows is None:
            active_rows = np.arange(self.n_rows)
        if active_cols is None:
            active_cols = np.arange(self.n_cols)
        self.iter_inds = []
        for row in active_rows:
            for col in active_cols:
                self.iter_inds.append([row, col])
        self.iter_inds = np.array(self.iter_inds)

        if self.simulation_mode:
            self.fm.surface.rwl = np.array([float(x) for x in self.infiles['reflectance_file'].meta['wavelength']])

    def get_components_at_index(self, index: int) -> (int, int, np.array, Geometry):
        """
        Get the spectrum from the file at the specified index.  Helper/
        parallel enabling function.

        :param index: reference location for iter_inds

        :return: success: boolean flag indicating if data present
        :return: r: row index
        :return: c: column index
        :return: meas: measured radiance file
        :return: geom: set up specified geometry files
        """
        # Determine the appropriate row, column index. and initialize the
        # data dictionary with empty entries.
        r, c = self.iter_inds[index]
        data = dict([(i, None) for i in self.input.get_all_element_names()])
        logging.debug('Row %i Column %i' % (r, c))

        # Read data from any of the input files that are defined.
        for source in self.infiles:
            data[source] = self.infiles[source].read_spectrum(r, c)
            if (index % flush_rate) == 0:
                self.infiles[source].flush_buffers()

        # Check for any bad data flags
        for source in self.infiles:
            if np.all(abs(data[source] - self.infiles[source].flag) < eps):
                return False, r, c, None, None

        if self.simulation_mode:
            # If solving the inverse problem, the measurment is the surface reflectance
            self.fm.surface.rfl = data['reflectance_file'].copy()
            if self.fm.surface.wl is not None:
                self.fm.surface.resample_reflectance()
            meas = self.fm.surface.rfl
        else:
            # If solving the inverse problem, the measurment is the radiance
            # We apply the calibration correciton here for simplicity.
            meas = data['measured_radiance_file']
            if meas is not None:
                meas = meas.copy()
            if data["radiometry_correction_file"] is not None:
                meas *= data['radiometry_correction_file']

        # We build the geometry object for this spectrum.  For files not
        # specified in the input configuration block, the associated entries
        # will be 'None'. The Geometry object will use reasonable defaults.
        geom = Geometry(obs=data['obs_file'],
                        glt=data['glt_file'],
                        loc=data['loc_file'],
                        bg_rfl=data['background_reflectance_file'])

        return True, r, c, meas, geom

    def __iter__(self):
        """ Reset the iterator"""

        self.iter = 0
        return self

    def __next__(self):
        """ Get the next spectrum from the file.  Turn the iteration number
            into row/column indices and read from all input products."""

        # Try to read data until we hit the end or find good values
        success = False
        while not success:
            if self.iter == len(self.iter_inds):
                self.flush_buffers()
                raise StopIteration

            # Determine the appropriate row, column index. and initialize the
            # data dictionary with empty entries.
            success, r, c, meas, geom = self.get_components_at_index(
                self.iter)
            self.iter = self.iter + 1

        return r, c, meas, geom

    def check_wavelengths(self, wl):
        """Make sure an input wavelengths align to the instrument definition."""

        return (len(wl) == self.fm.instrument.wl) and \
            all((wl-self.fm.instrument.wl) < 1e-2)

    def flush_buffers(self):
        """Write all buffered output data to disk, and erase read buffers."""

        for file_dictionary in [self.infiles, self.outfiles]:
            for name, fi in file_dictionary.items():
                fi.flush_buffers()

    def write_spectrum(self, row, col, states, meas, geom, flush_immediately=False):
        """Write data from a single inversion to all output buffers."""

        self.writes = self.writes + 1

        if len(states) == 0:

            # Write a bad data flag
            atm_bad = np.zeros(len(self.fm.instrument.n_chan)*5) * -9999.0
            state_bad = np.zeros(len(self.fm.statevec)) * -9999.0
            data_bad = np.zeros(self.fm.instrument.n_chan) * -9999.0
            to_write = {
                'estimated_state_file': state_bad,
                'estimated_reflectance_file': data_bad,
                'estimated_emission_file': data_bad,
                'modeled_radiance_file': data_bad,
                'apparent_reflectance_file': data_bad,
                'path_radiance_file': data_bad,
                'simulated_measurement_file': data_bad,
                'algebraic_inverse_file': data_bad,
                'atmospheric_coefficients_file': atm_bad,
                'radiometry_correction_file': data_bad,
                'spectral_calibration_file': data_bad,
                'posterior_uncertainty_file': state_bad
            }

        else:

            # The inversion returns a list of states, which are
            # intepreted either as samples from the posterior (MCMC case)
            # or as a gradient descent trajectory (standard case). For
            # gradient descent the last spectrum is the converged solution.
            if self.iv.mode == 'mcmc':
                state_est = states.mean(axis=0)
            else:
                state_est = states[-1, :]

            # Spectral calibration
            wl, fwhm = self.fm.calibration(state_est)
            cal = np.column_stack(
                [np.arange(0, len(wl)), wl / 1000.0, fwhm / 1000.0])

            # If there is no actual measurement, we use the simulated version
            # in subsequent calculations.  Naturally in these cases we're
            # mostly just interested in the simulation result.
            if meas is None:
                meas = self.fm.calc_rdn(state_est, geom)

            # Rodgers diagnostics
            lamb_est, meas_est, path_est, S_hat, K, G = \
                self.iv.forward_uncertainty(state_est, meas, geom)

            # Simulation with noise
            meas_sim = self.fm.instrument.simulate_measurement(meas_est, geom)

            # Algebraic inverse and atmospheric optical coefficients
            x_surface, x_RT, x_instrument = self.fm.unpack(state_est)
            rfl_alg_opt, Ls, coeffs = invert_algebraic(self.fm.surface,
                                                       self.fm.RT, self.fm.instrument, x_surface, x_RT, x_instrument,
                                                       meas, geom)
            rhoatm, sphalb, transm, solar_irr, coszen, transup = coeffs

            L_atm = self.fm.RT.get_L_atm(x_RT, geom)
            L_down_transmitted = self.fm.RT.get_L_down_transmitted(x_RT, geom)

            atm = np.column_stack(list(coeffs[:4]) +
                                  [np.ones((len(wl), 1)) * coszen])
            atm = atm.T.reshape((len(wl)*5,))

            # Upward emission & glint and apparent reflectance
            Ls_est = self.fm.calc_Ls(state_est, geom)
            apparent_rfl_est = lamb_est + Ls_est

            # Radiometric calibration
            factors = np.ones(len(wl))
            if 'radiometry_correction_file' in self.outfiles:
                if 'reference_reflectance_file' in self.infiles:
                    reference_file = self.infiles['reference_reflectance_file']
                    self.rfl_ref = reference_file.read_spectrum(row, col)
                    self.wl_ref = reference_file.wl
                    w, fw = self.fm.instrument.calibration(x_instrument)
                    resamp = resample_spectrum(self.rfl_ref, self.wl_ref,
                                               w, fw, fill=True)
                    meas_est = self.fm.calc_meas(state_est, geom, rfl=resamp)
                    factors = meas_est / meas
                else:
                    logging.warning('No reflectance reference')

            # Assemble all output products
            to_write = {
                'estimated_state_file': state_est,
                'estimated_reflectance_file': np.column_stack((self.fm.surface.wl, lamb_est)),
                'estimated_emission_file': np.column_stack((self.fm.surface.wl, Ls_est)),
                'modeled_radiance_file': np.column_stack((wl, meas_est)),
                'apparent_reflectance_file': np.column_stack((self.fm.surface.wl, apparent_rfl_est)),
                'path_radiance_file': np.column_stack((wl, path_est)),
                'simulated_measurement_file': np.column_stack((wl, meas_sim)),
                'algebraic_inverse_file': np.column_stack((self.fm.surface.wl, rfl_alg_opt)),
                'atmospheric_coefficients_file': atm,
                'radiometry_correction_file': factors,
                'spectral_calibration_file': cal,
                'posterior_uncertainty_file': np.sqrt(np.diag(S_hat))
            }

        for product in self.outfiles:
            logging.debug('IO: Writing '+product)
            self.outfiles[product].write_spectrum(row, col, to_write[product])
            if (self.writes % flush_rate) == 0 or flush_immediately:
                self.outfiles[product].flush_buffers()

        # Special case! samples file is matlab format.
        if self.output.mcmc_samples_file is not None:
            logging.debug('IO: Writing mcmc_samples_file')
            mdict = {'samples': states}
            scipy.io.savemat(self.output.mcmc_samples_file, mdict)

        # Special case! Data dump file is matlab format.
        if self.output.data_dump_file is not None:

            logging.debug('IO: Writing data_dump_file')
            x = state_est
            xall = states
            Seps_inv, Seps_inv_sqrt = self.iv.calc_Seps(x, meas, geom)
            meas_est_window = meas_est[self.iv.winidx]
            meas_window = meas[self.iv.winidx]
            xa, Sa, Sa_inv, Sa_inv_sqrt = self.iv.calc_prior(x, geom)
            prior_resid = (x - xa).dot(Sa_inv_sqrt)
            rdn_est = self.fm.calc_rdn(x, geom)
            rdn_est_all = np.array([self.fm.calc_rdn(xtemp, geom)
                                    for xtemp in states])

            x_surface, x_RT, x_instrument = self.fm.unpack(x)
            Kb = self.fm.Kb(x, geom)
            xinit = invert_simple(self.fm, meas, geom)
            Sy = self.fm.instrument.Sy(meas, geom)
            cost_jac_prior = np.diagflat(x - xa).dot(Sa_inv_sqrt)
            cost_jac_meas = Seps_inv_sqrt.dot(K[self.iv.winidx, :])
            meas_Cov = self.fm.Seps(x, meas, geom)
            lamb_est, meas_est, path_est, S_hat, K, G = \
                self.iv.forward_uncertainty(state_est, meas, geom)
            A = np.matmul(K, G)

            # Form the MATLAB dictionary object and write to file
            mdict = {
                'K': K,
                'G': G,
                'S_hat': S_hat,
                'prior_mu': xa,
                'Ls': Ls,
                'prior_Cov': Sa,
                'meas': meas,
                'rdn_est': rdn_est,
                'rdn_est_all': rdn_est_all,
                'x': x,
                'xall': xall,
                'x_surface': x_surface,
                'x_RT': x_RT,
                'x_instrument': x_instrument,
                'meas_Cov': meas_Cov,
                'wl': wl,
                'fwhm': fwhm,
                'lamb_est': lamb_est,
                'coszen': coszen,
                'cost_jac_prior': cost_jac_prior,
                'Kb': Kb,
                'A': A,
                'cost_jac_meas': cost_jac_meas,
                'winidx': self.iv.winidx,
                'windows': self.iv.windows,
                'prior_resid': prior_resid,
                'noise_Cov': Sy,
                'xinit': xinit,
                'rhoatm': rhoatm,
                'sphalb': sphalb,
                'transm': transm,
                'transup': transup,
                'solar_irr': solar_irr,
                'L_atm': L_atm,
                'L_down_transmitted': L_down_transmitted
            }
            scipy.io.savemat(self.output.data_dump_file, mdict)

        # Write plots, if needed
        if len(states) > 0 and self.output.plot_directory is not None:

            if 'reference_reflectance_file' in self.infiles:
                reference_file = self.infiles['reference_reflectance_file']
                self.rfl_ref = reference_file.read_spectrum(row, col)
                self.wl_ref = reference_file.wl

            for i, x in enumerate(states):

                # Calculate intermediate solutions
                lamb_est, meas_est, path_est, S_hat, K, G = \
                    self.iv.forward_uncertainty(state_est, meas, geom)

                plt.cla()
                red = [0.7, 0.2, 0.2]
                wl, fwhm = self.fm.calibration(x)
                xmin, xmax = min(wl), max(wl)
                fig = plt.subplots(1, 2, figsize=(10, 5))
                plt.subplot(1, 2, 1)
                meas_est = self.fm.calc_meas(x, geom)
                for lo, hi in self.iv.windows:
                    idx = np.where(np.logical_and(wl > lo, wl < hi))[0]
                    p1 = plt.plot(wl[idx], meas[idx], color=red, linewidth=2)
                    p2 = plt.plot(wl, meas_est, color='k', linewidth=1)
                plt.title("Radiance")
                plt.title("Measurement (Scaled DN)")
                ymax = max(meas)*1.25
                ymax = max(meas)+0.01
                ymin = min(meas)-0.01
                plt.text(500, ymax*0.92, "Measured", color=red)
                plt.text(500, ymax*0.86, "Model", color='k')
                plt.ylabel(r"$\mu$W nm$^{-1}$ sr$^{-1}$ cm$^{-2}$")
                plt.ylabel("Intensity")
                plt.xlabel("Wavelength (nm)")
                plt.ylim([-0.001, ymax])
                plt.ylim([ymin, ymax])
                plt.xlim([xmin, xmax])

                plt.subplot(1, 2, 2)
                lamb_est = self.fm.calc_lamb(x, geom)
                ymax = min(max(lamb_est)*1.25, 0.10)
                for lo, hi in self.iv.windows:

                    # black line
                    idx = np.where(np.logical_and(wl > lo, wl < hi))[0]
                    p2 = plt.plot(wl[idx], lamb_est[idx], 'k', linewidth=2)
                    ymax = max(max(lamb_est[idx]*1.2), ymax)

                    # red line
                    if 'reference_reflectance_file' in self.infiles:
                        idx = np.where(np.logical_and(
                            self.wl_ref > lo, self.wl_ref < hi))[0]
                        p1 = plt.plot(self.wl_ref[idx], self.rfl_ref[idx],
                                      color=red, linewidth=2)
                        ymax = max(max(self.rfl_ref[idx]*1.2), ymax)

                    # green and blue lines - surface components
                    if hasattr(self.fm.surface, 'components') and \
                            self.output.plot_surface_components:
                        idx = np.where(np.logical_and(self.fm.surface.wl > lo,
                                                      self.fm.surface.wl < hi))[0]
                        p3 = plt.plot(self.fm.surface.wl[idx],
                                      self.fm.xa(x, geom)[idx], 'b', linewidth=2)
                        for j in range(len(self.fm.surface.components)):
                            z = self.fm.surface.norm(
                                lamb_est[self.fm.surface.idx_ref])
                            mu = self.fm.surface.components[j][0] * z
                            plt.plot(self.fm.surface.wl[idx], mu[idx], 'g:',
                                     linewidth=1)
                plt.text(500, ymax*0.86, "Remote estimate", color='k')
                if 'reference_reflectance_file' in self.infiles:
                    plt.text(500, ymax*0.92, "In situ reference", color=red)
                if hasattr(self.fm.surface, 'components') and \
                        self.output.plot_surface_components:
                    plt.text(500, ymax*0.80, "Prior mean state ",
                             color='b')
                    plt.text(500, ymax*0.74, "Surface components ",
                             color='g')
                plt.ylim([-0.0010, ymax])
                plt.xlim([xmin, xmax])
                plt.title("Reflectance")
                plt.title("Source Model")
                plt.xlabel("Wavelength (nm)")
                fn = os.path.join(self.output.plot_directory, ('frame_%i.png' % i))
                plt.savefig(fn)
                plt.close()



#! /usr/bin/env python
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#         Adam Erickson, adam.m.erickson@nasa.gov
#


### Variables ###

name = 'isofit'

__version__ = '2.8.0'

warnings_enabled = False




from isofit.core.geometry import Geometry


def test_Geometry():
    geom = Geometry()
    assert geom.earth_sun_file == None
    assert geom.observer_zenith == 0
    assert geom.observer_azimuth == 0
    assert geom.observer_altitude_km == None
    assert geom.surface_elevation_km == None
    assert geom.datetime == None
    assert geom.day_of_year == None
    assert geom.latitude == None
    assert geom.longitude == None
    assert geom.longitudeE == None
    assert geom.gmtime == None
    assert geom.earth_sun_distance == None
    assert geom.OBSZEN == 180.0
    assert geom.RELAZ == 0.0
    assert geom.TRUEAZ == 0.0
    assert geom.umu == 1.0



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: Philip G. Brodrick, philip.brodrick@jpl.nasa.gov

import logging
from collections import OrderedDict
from typing import Dict, List, Type
import os
from isofit.configs.sections.input_config import InputConfig
from isofit.configs.sections.output_config import OutputConfig
from isofit.configs.sections.forward_model_config import ForwardModelConfig
from isofit.configs.sections.implementation_config import ImplementationConfig
from isofit.configs.base_config import BaseConfigSection
from isofit.core import common
import yaml


class Config(BaseConfigSection):
    """
    Handles the reading and formatting of configuration files.  Please note - there are many ways to do this, some
    of which require fewer lines of code.  This method was chosen to facilitate more clarity when using / adding /
    modifying code, particularly given the highly flexible nature of Isofit.

    How to use:

    To add an additional parameter to an existing class, simply go to the relevant config (e.g. for forward_model go
    to sections/forward_model_config.py), and in the config class (e.g. ForwardModelConfig) add the parameter.  Also
    Add a hidden parameter with the _type suffix, which will be used to check that configs read the appropriate type.
    Add comments directly below, to be auto-appended to online documentation.
    Example::
        class GenericConfigSection(BaseConfigSection):
            _attribute_type = str
            attribute = 'my attribute'
            \"""str: attribute does whatever it happens to do\"""

    To validate that attributes have appropriate relationships or characteristics, use the hidden _check_config_validity
    method to add more detailed validation checks. Simply return a list of string descriptions of errors from the
    method as demonstrated::
        def _check_config_validity(self) -> List[str]:
            errors = list()
            if self.attribute_min >= self.attribute_max:
                errors.append('attribute_min must be less than attribute_max.')
            return errors

    """

    def __init__(self, configdict) -> None:

        self._input_type = InputConfig
        self.input = InputConfig({})
        """InputConfig: Input config. Holds all input file information.
        """

        self._output_type = OutputConfig
        self.output = OutputConfig({})
        """OutputConfig: Output config. Holds all output file information.
        """

        self._forward_model_type = ForwardModelConfig
        self.forward_model = ForwardModelConfig({})
        """ForwardModelConfig: forward_model config. Holds information about surface models, 
        radiative transfer models, and the instrument.
        """

        self._implementation_type = ImplementationConfig
        self.implementation = ImplementationConfig({})
        """ImplementationConfig: holds information regarding how isofit is to be run, including relevant sub-configs 
        (e.g. inversion information).
        """

        # Load sub-classes and attributes
        self.set_config_options(configdict)

    def get_config_as_dict(self) -> dict:
        """Get configuration options as a nested dictionary with delineated sections.

        Returns:
            Configuration options as a nested dictionary with delineated sections.
        """
        config = OrderedDict()
        for config_section in self._get_nontype_attributes():
            populated_section = getattr(self, config_section)
            config[config_section] = populated_section.get_config_options_as_dict()
        return config

    def get_config_errors(self):
        """
        Get configuration option errors by checking the validity of each config section.
        """
        logging.info("Checking config sections for configuration issues")

        errors = self.check_config_validity()

        for e in errors:
            logging.error(e)

        if len(errors) > 0:
            raise AttributeError('Configuration error(s) found.  See log for details.')

        logging.info('Configuration file checks complete, no errors found.')


def get_config_differences(config_a: Config, config_b: Config) -> Dict:
    differing_items = dict()
    dict_a = config_a.get_config_as_dict()
    dict_b = config_b.get_config_as_dict()
    all_sections = set(list(dict_a.keys()) + list(dict_b.keys()))
    for section in all_sections:
        section_a = dict_a.get(section, dict())
        section_b = dict_b.get(section, dict())
        all_options = set(list(section_a.keys()) + list(section_b.keys()))
        for option in all_options:
            value_a = section_a.get(option, None)
            value_b = section_b.get(option, None)
            if value_a != value_b:
                logging.debug(
                    "Configs have different values for option {} in section {}:  {} and {}".format(
                        option, section, value_a, value_b
                    )
                )
                differing_items.setdefault(section, dict())[option] = (value_a, value_b)
    return differing_items


def create_new_config(config_file: str) -> Config:
    """Load a config file from disk.
    Args:
        config_file: file to load config from.  Currently accepted formats: JSON and YAML

    Returns:
        Config object, having completed all necessary config checks
    """
    try:
        with open(config_file, 'r') as f:
            config_dict = yaml.safe_load(f)
    except:
        raise IOError('Unexpected configuration file time, only json and yaml supported')

    configdir, f = os.path.split(os.path.abspath(config_file))

    config_dict = common.expand_all_paths(config_dict, configdir)
    config = Config(config_dict)

    return config



#! /usr/bin/env python3
#
#  Copyright 2019 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import numpy as np
from spectral.io import envi
import ray
import ray.services
import logging
import atexit

@ray.remote
def extract_chunk(lstart: int, lend: int, in_file: str, labels, flag, logfile=None, loglevel='INFO'):
    """
    Extract a small chunk of the image

    Args:
        lstart: line to start extraction at
        lend: line to end extraction at
        in_file: file to read image from
        label: labels to use for data read
        flag: nodata value of image
        logfile: logging file name
        loglevel: logging level

    Returns:
        out_index: array of output indices (based on labels)
        out_data: array of output data
    """
        
    logging.basicConfig(format='%(levelname)s:%(message)s', level=loglevel, filename=logfile)
    logging.info(f'{lstart}: starting')

    in_img = envi.open(in_file + '.hdr')
    img_mm = in_img.open_memmap(interleave='bip', writable=False)

    # Which labels will we extract? ignore zero index
    active = labels[lstart:lend, :]
    active = active[active >= 1]
    active = np.unique(active)
    logging.debug(f'{lstart}: found {len(active)} unique labels')
    if len(active) == 0:
        return None, None

    # Handle labels extending outside our chunk by expanding margins
    cs = lend - lstart
    boundary_min = max(lstart-cs, 0)
    boundary_max = min(lend+cs, labels.shape[0])

    active_area = np.zeros((boundary_max-boundary_min,labels.shape[1]))
    for i in active:
        active_area[labels[boundary_min:boundary_max,:] == i] = True
    active_locs = np.where(active_area)

    lstart_adjust = min(active_locs[0]) + boundary_min
    lend_adjust = max(active_locs[0]) + boundary_min + 1

    cstart_adjust = min(active_locs[1])
    cend_adjust = max(active_locs[1]) + 1

    logging.debug(f'{lstart} area subset: {lstart_adjust}, {lend_adjust} :::: {cstart_adjust}, {cend_adjust}')

    chunk_lbl = np.array(labels[lstart_adjust:lend_adjust, cstart_adjust:cend_adjust])
    chunk_inp = np.array(img_mm[lstart_adjust:lend_adjust, cstart_adjust:cend_adjust, :])

    out_data = np.zeros((len(active),img_mm.shape[-1])) + flag

    logging.debug(f'{lstart}: running extraction from local array')
    for _lab, lab in enumerate(active):
        out_data[_lab, :] = 0
        locs = np.where(chunk_lbl == lab)
        for row, col in zip(locs[0], locs[1]):
            out_data[_lab, :] += np.squeeze(chunk_inp[row, col, :])
        out_data[_lab,:] /= float(len(locs[0]))

    logging.debug(f'{lstart}: complete')
    return active.astype(int), out_data




def extractions(inputfile, labels, output, chunksize, flag, n_cores: int = 1, ray_address: str = None, ray_redis_password: str = None, ray_temp_dir: str = None, ray_ip_head = None, logfile: str = None, loglevel: str = 'INFO'):
    """..."""

    in_file = inputfile
    lbl_file = labels
    out_file = output
    nchunk = chunksize

    dtm = {
        '4': np.float32,
        '5': np.float64
    }

    # Open input data, get dimensions
    in_img = envi.open(in_file+'.hdr', in_file)
    meta = in_img.metadata

    nl, nb, ns = [int(meta[n]) for n in ('lines', 'bands', 'samples')]
    img_mm = in_img.open_memmap(interleave='bip', writable=False)

    lbl_img = envi.open(lbl_file+'.hdr', lbl_file)
    labels = lbl_img.read_band(0)
    un_labels = np.unique(labels).tolist()
    if 0 not in un_labels:
        un_labels.insert(0,0)
    nout = len(un_labels)


    # Start up a ray instance for parallel work
    rayargs = {'ignore_reinit_error': True,
               'local_mode': n_cores == 1,
               "address": ray_address,
               "_redis_password": ray_redis_password}

    if rayargs['local_mode']:
        rayargs['_temp_dir'] = ray_temp_dir
        # Used to run on a VPN
        ray.services.get_node_ip_address = lambda: '127.0.0.1'

    # We can only set the num_cpus if running on a single-node
    if ray_ip_head is None and ray_redis_password is None:
        rayargs['num_cpus'] = n_cores

    ray.init(**rayargs)
    atexit.register(ray.shutdown)

    labelid = ray.put(labels)
    jobs = []
    for lstart in np.arange(0, nl, nchunk):
        lend = min(lstart+nchunk, nl)
        jobs.append(extract_chunk.remote(lstart, lend, in_file, labelid, flag, logfile=logfile, loglevel=loglevel))


    # Collect results
    rreturn = [ray.get(jid) for jid in jobs]

    # Iterate through image "chunks," segmenting as we go
    out = np.zeros((nout, nb))
    for idx, ret in rreturn:
        if ret is not None:
            for _output_index in range(len(idx)):
                out[un_labels.index(idx[_output_index]),...] = ret[_output_index,...]
    del rreturn
    ray.shutdown()

    out[np.logical_not(np.isfinite(out))] = flag

    meta["lines"] = str(nout)
    meta["bands"] = str(nb)
    meta["samples"] = '1'
    meta["interleave"] = "bil"

    out_img = envi.create_image(out_file+'.hdr',  metadata=meta,
                                ext='', force=True)
    out_mm = np.memmap(out_file, dtype=dtm[meta['data type']], mode='w+',
                      shape=(nout, 1, nb))
    if dtm[meta['data type']] == np.float32:
        out_mm[:, 0, :] = np.array(out, np.float32)
    else:
        out_mm[:, 0, :] = np.array(out, np.float64)



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: Philip G. Brodrick, philip.brodrick@jpl.nasa.gov

from typing import Dict, List, Type
from isofit.configs.base_config import BaseConfigSection
import os


class InputConfig(BaseConfigSection):

    def __init__(self, sub_configdic: dict = None):
        """
        Input file(s) configuration.
        """

        self._measured_radiance_file_type = str
        self.measured_radiance_file = None
        """
        str: Input radiance file.  Can be either a .mat, .txt, or ENVI formatted binary cube. 
        Used for inverse-modeling (radiance -> reflectance).
        """

        self._reference_reflectance_file_type = str
        self.reference_reflectance_file = None
        """
        str: Input reference reflectance file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        Used for radiometric calibration.
        """

        self._reflectance_file_type = str
        self.reflectance_file = None
        """
        str: Input reflectance file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        Used for forward-modeling (reflectance -> radiance).
        """

        self._obs_file_type = str
        self.obs_file = None
        """
        str: Input 'obs', or observation, file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        Provides information about the conditions during observaiton.  Assumed to be in the band-wise format:
        {path length, to-sensor azimuth, to-sensor zenith, to-sun azimuth, to-sun zenith, phase, slope, aspect, cosine i, 
        UTC time}
        """

        self._glt_file_type = str
        self.glt_file = None
        """
        str: Input glt file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        Provides (x,y) offset information for the spatial location of raw-space input files
        """

        self._loc_file_type = str
        self.loc_file = None
        """
        str: Input 'loc', or location, file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        Provides per-pixel lat, long, and elevation information.
        """
                
        self._background_reflectance_file_type = str
        self.background_reflectance_file = None
        """
        str: Background reflectance file used for spatial inference.
        Should be an ENVI formatted binary cube that is the same
        dimensions as the input data cube.
        """

        self._surface_prior_mean_file_type = str
        self.surface_prior_mean_file = None
        """
        str: Input surface prior mean file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        """

        self._surface_prior_variance_file_type = str
        self.surface_prior_variance_file = None
        """
        str: Input surface prior variance file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        """

        self._rt_prior_mean_file_type = str
        self.rt_prior_mean_file = None
        """
        str: Input rt prior mean file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        """

        self._rt_prior_variance_file_type = str
        self.rt_prior_variance_file = None
        """
        str: Input rt prior variance file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        """

        self._instrument_prior_mean_file_type = str
        self.instrument_prior_mean_file = None
        """
        str: Input instrument prior mean file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        """

        self._instrument_prior_variance_file_type = str
        self.instrument_prior_variance_file = None
        """
        str: Input instrument prior variance file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        """

        self._radiometry_correction_file_type = str
        self.radiometry_correction_file = None
        """
        str: Input radiometric correction file.  Can be either a .mat, .txt, or ENVI formatted binary cube.
        Used to make minor channelized corrections to account for slight systematic errors not captured in calibration.
        """

        self.set_config_options(sub_configdic)

    def _check_config_validity(self) -> List[str]:
        errors = list()

        # Check that all input files exist
        for key in self._get_nontype_attributes():
            value = getattr(self, key)
            if value is not None:
                if os.path.isfile(value) is False:
                    errors.append('Config value Input->{}: {} not found'.format(key, value))

        # TODO: check that the right combination of input files exists

        # Recursive call to any sub-config-module errors
        for key in self.__dict__.keys():
            value = getattr(self, key)
            if callable(key):
                errors.extend(value.check_config_validity())

        return errors



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import scipy as s

from ..core.common import eps
from .surface_multicomp import MultiComponentSurface
from .surface_thermal import ThermalSurface
from isofit.configs import Config


class GlintSurface(ThermalSurface):
    """A model of the surface based on a collection of multivariate 
       Gaussians, extended with a surface glint term."""

    def __init__(self, full_config: Config):

        super().__init__(full_config)

        # TODO: Enforce this attribute in the config, not here (this is hidden)
        self.statevec_names.extend(['GLINT'])
        self.scale.extend([1.0])
        self.init.extend([0.005])
        self.bounds.extend([[0, 0.2]])
        self.n_state = self.n_state + 1
        self.glint_ind = len(self.statevec_names) - 1

    def xa(self, x_surface, geom):
        """Mean of prior distribution, calculated at state x."""

        mu = ThermalSurface.xa(self, x_surface, geom)
        mu[self.glint_ind] = self.init[self.glint_ind]
        return mu

    def Sa(self, x_surface, geom):
        """Covariance of prior distribution, calculated at state x.  We find
        the covariance in a normalized space (normalizing by z) and then un-
        normalize the result for the calling function."""

        Cov = ThermalSurface.Sa(self, x_surface, geom)
        f = s.array([[(10.0 * self.scale[self.glint_ind])**2]])
        Cov[self.glint_ind, self.glint_ind] = f
        return Cov

    def fit_params(self, rfl_meas, geom, *args):
        """Given a reflectance estimate and one or more emissive parameters, 
          fit a state vector."""

        glint_band = s.argmin(abs(900-self.wl))
        glint = s.mean(rfl_meas[(glint_band-2):glint_band+2])
        water_band = s.argmin(abs(400-self.wl))
        water = s.mean(rfl_meas[(water_band-2):water_band+2])
        if glint > 0.05 or water < glint:
            glint = 0
        glint = max(self.bounds[self.glint_ind][0]+eps,
                    min(self.bounds[self.glint_ind][1]-eps, glint))
        lamb_est = rfl_meas - glint
        x = ThermalSurface.fit_params(self, lamb_est, geom)
        x[self.glint_ind] = glint
        return x

    def calc_rfl(self, x_surface, geom):
        """Reflectance (includes specular glint)."""

        return self.calc_lamb(x_surface, geom) + x_surface[self.glint_ind]

    def drfl_dsurface(self, x_surface, geom):
        """Partial derivative of reflectance with respect to state vector, 
        calculated at x_surface."""

        drfl = self.dlamb_dsurface(x_surface, geom)
        drfl[:, self.glint_ind] = 1
        return drfl

    def dLs_dsurface(self, x_surface, geom):
        """Partial derivative of surface emission with respect to state vector, 
        calculated at x_surface.  We append a column of zeros to handle 
        the extra glint parameter"""

        dLs_dsurface = super().dLs_dsurface(x_surface, geom)
        dLs_dglint = s.zeros((dLs_dsurface.shape[0],1))
        dLs_dsurface = s.hstack([dLs_dsurface, dLs_dglint]) 
        return dLs_dsurface

    def summarize(self, x_surface, geom):
        """Summary of state vector."""

        return ThermalSurface.summarize(self, x_surface, geom) + \
            ' Glint: %5.3f' % x_surface[self.glint_ind]



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: Philip G. Brodrick, philip.brodrick@jpl.nasa.gov

from typing import Dict, List, Type
from isofit.configs.base_config import BaseConfigSection
from isofit.configs.sections.instrument_config import InstrumentConfig
from isofit.configs.sections.surface_config import SurfaceConfig
from isofit.configs.sections.radiative_transfer_config import RadiativeTransferConfig


class ForwardModelConfig(BaseConfigSection):
    """
    Forward model configuration.
    """

    def __init__(self, sub_configdic: dict = None):
        self._instrument_type = InstrumentConfig
        self.instrument: InstrumentConfig = None
        """
        Instrument: instrument config section. 
        """

        self._surface_type = SurfaceConfig
        self.surface: SurfaceConfig = None
        """
        Instrument: instrument config section. 
        """

        self._radiative_transfer_type = RadiativeTransferConfig
        self.radiative_transfer: RadiativeTransferConfig = None
        """
        RadiativeTransfer: radiative transfer config section.
        """

        self._model_discrepancy_file_type = str
        self.model_discrepancy_file = None
        """
        Points to an numpy-format covariance matrix. 
        """

        self.set_config_options(sub_configdic)

    def _check_config_validity(self) -> List[str]:
        errors = list()

        return errors



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import os
import numpy as np
import logging
import ray
from collections import OrderedDict
import subprocess
import time
import atexit

from isofit.core import common
from isofit.configs import Config
from isofit.configs.sections.radiative_transfer_config import RadiativeTransferEngineConfig
from isofit.configs.sections.statevector_config import StateVectorElementConfig
from isofit.configs.sections.implementation_config import ImplementationConfig


### Functions ###

@ray.remote
def spawn_rt(cmd, local_dir=None):
    """Run a CLI command."""

    print(cmd)

    # Add a very slight timing offset to prevent all subprocesses
    # starting simultaneously
    time.sleep(float(np.random.random(1))*2)

    subprocess.call(cmd, shell=True, cwd=local_dir)

### Classes ###

class FileExistsError(Exception):
    """FileExistsError with a message."""

    def __init__(self, message):
        super(FileExistsError, self).__init__(message)


class TabularRT:
    """A model of photon transport including the atmosphere."""

    def __init__(self, engine_config: RadiativeTransferEngineConfig, full_config: Config):

        self.implementation_config: ImplementationConfig = full_config.implementation
        self.wl, self.fwhm = common.load_wavelen(full_config.forward_model.instrument.wavelength_file)
        if engine_config.wavelength_range is not None:
            valid_wl = np.logical_and(self.wl >= engine_config.wavelength_range[0],
                                      self.wl <= engine_config.wavelength_range[1])
            self.wl = self.wl[valid_wl]
            self.fwhm = self.fwhm[valid_wl]

        self.n_chan = len(self.wl)

        self.auto_rebuild = full_config.implementation.rte_auto_rebuild
        self.configure_and_exit = full_config.implementation.rte_configure_and_exit

        # We use a sorted dictionary here so that filenames for lookup
        # table (LUT) grid points are always constructed the same way, with
        # consistent dimesion ordering). Every state vector element has
        # a lookup table dimension, but some lookup table dimensions
        # (like geometry parameters) may not be in the state vector.
        # TODO: enforce a requirement that makes all SV elements be inside the LUT
        full_lut_grid = full_config.forward_model.radiative_transfer.lut_grid
        # selectively get lut components that are in this particular RTE
        self.lut_grid_config = OrderedDict()
        if engine_config.lut_names is not None:
            lut_names = engine_config.lut_names
        else:
            lut_names = full_config.forward_model.radiative_transfer.lut_grid.keys()

        for key, value in full_lut_grid.items():
            if key in lut_names:
                self.lut_grid_config[key] = value

        # selectively get statevector components that are in this particular RTE
        full_sv_names = full_config.forward_model.radiative_transfer.statevector.get_element_names()
        self.statevector_names = full_sv_names

        self.lut_dir = engine_config.lut_path
        self.n_point = len(self.lut_grid_config)
        self.n_state = len(self.statevector_names)

        self.luts = {}

        # Retrieved variables.  We establish scaling, bounds, and
        # initial guesses for each state vector element.  The state
        # vector elements are all free parameters in the RT lookup table,
        # and they all have associated dimensions in the LUT grid.
        self.bounds, self.scale, self.init = [], [], []
        self.prior_mean, self.prior_sigma = [], []
        for key in self.statevector_names:
            element: StateVectorElementConfig = full_config.forward_model.radiative_transfer.statevector.get_single_element_by_name(
                key)
            self.bounds.append(element.bounds)
            self.scale.append(element.scale)
            self.init.append(element.init)
            self.prior_sigma.append(element.prior_sigma)
            self.prior_mean.append(element.prior_mean)
        self.bounds = np.array(self.bounds)
        self.scale = np.array(self.scale)
        self.init = np.array(self.init)
        self.prior_mean = np.array(self.prior_mean)
        self.prior_sigma = np.array(self.prior_sigma)

        self.lut_dims = []
        self.lut_grids = []
        self.lut_names = []
        self.lut_interp_types = []
        for key, grid_values in self.lut_grid_config.items():

            # do some quick checks on the values
            if len(grid_values) == 1:
                err = 'Only 1 value in LUT grid {}. ' +\
                    '1-d LUT grids cannot be interpreted.'.format(key)
                raise ValueError(err)
            if grid_values != sorted(grid_values):
                logging.error('Lookup table grid needs ascending order')
                raise ValueError('Lookup table grid needs ascending order')

            # Store the values
            self.lut_grids.append(grid_values)
            self.lut_dims.append(len(grid_values))
            self.lut_names.append(key)

            # Store in an indication of the type of value each key is
            # (normal - n, degree - d, radian - r)
            if key in self.angular_lut_keys_radians:
                self.lut_interp_types.append('r')
            elif key in self.angular_lut_keys_degrees:
                self.lut_interp_types.append('d')
            else:
                self.lut_interp_types.append('n')

        # Cast as array for faster reference later
        self.lut_interp_types = np.array(self.lut_interp_types)

        # "points" contains all combinations of grid points
        # We will have one filename prefix per point
        self.points = common.combos(self.lut_grids)
        self.files = self.get_lut_filenames()

    def build_lut(self, rebuild=False):
        """Each LUT is associated with a source directory.  We build a lookup 
            table by: 
              (1) defining the LUT dimensions, state vector names, and the 
                    grid of values; 
              (2) running the radiative transfer solver if needed, with each 
                    run defining a different point in the LUT; and 
              (3) loading the LUTs, one per key atmospheric coefficient vector,
                  into memory as VectorInterpolator objects."""

       # Build the list of radiative transfer run commands. This
        # rebuild_cmd() function will be overriden by the child class to
        # perform setup activities unique to each RTM.
        rebuild_cmds = []
        for point, fn in zip(self.points, self.files):
            try:
                cmd = self.rebuild_cmd(point, fn)
                rebuild_cmds.append(cmd)
            except FileExistsError:
                pass

        if self.configure_and_exit:
            raise SystemExit
            # sys.exit(0)

        elif len(rebuild_cmds) > 0 and self.auto_rebuild:
            logging.info("Rebuilding radiative transfer look up table")

            # check to make sure lut directory is there, create if not
            if os.path.isdir(self.lut_dir) is False:
                os.mkdir(self.lut_dir)

            # Make the LUT calls (in parallel if specified)
            results = ray.get([spawn_rt.remote(rebuild_cmd, self.lut_dir) for rebuild_cmd in rebuild_cmds])


    def get_lut_filenames(self):
        files = []
        for point in self.points:
            outf = '_'.join(['%s-%6.4f' % (n, x)
                             for n, x in zip(self.lut_names, point)])
            files.append(outf)
        return files

    def summarize(self, x_RT, geom):
        """Summary of state vector."""

        if len(x_RT) < 1:
            return ''
        return 'Atmosphere: '+' '.join(['%s: %5.3f' % (si, xi) for si, xi in
                                        zip(self.statevector_names, x_RT)])
